/******************************************************************************
 * File: pvm1.c
 * Data collection program for my Solar Power Inverter.
 * Idea is to send out pre-arranged sets of bytes to a remote
 * device at specified intervals.
 * The responses are recorded in a DOC-format text file, along with
 * timestamps.
 * Program will continue to schedule samples if the Palm goes to sleep,
 * or it is turned off by the power button.
 * If the user switches to another application, sampling is stopped.
 * DOC file routines leveraged from Zdocj editor, which is itself GPL.
 * 
 * To adapt this program for another device, look for the string, "Custom:"
 * for hints as to where changes are needed.
 * 
 * Started: 30-AUG-2007; Roderick Young.
 * 0.1	Initial version
 * 0.2	Reading successfully and displaying most values. Writing file.
 * 0.3	Calculating efficiency.  Adjustments to calculation constants.
 * 		Raw hex data dumped to facilitate debugging.
 * 1.0	First release.  Add safeguard so that the log file won't corrupt
 * 		and wrap around if it fills up with 65536 records.
 * 1.1	Power on serial port only while actually communicating with
 * 		inverter.  Save a little PDA power.
 *****************************************************************************/

#include <PalmOS.h>
#include <ByteOrderUtils.h>

#include "pvm1_c.h"
#include "pvm1_g.h"

// IDs for saved Preferences
#define PREF_VER 1
#define PREFID_BASE 0

// Custom: This is the format of the buffer returned by the inverter
typedef union {
	unsigned char buf[16];  // raw byte buffer
	struct {  // format of response to AA 01 command
		UInt16 checksum;
		UInt32 status;
		UInt16 pout;
		UInt16 vin;
		UInt16 iin;
		Int16 temp;
		UInt16 vout;
	} f1;
	struct {  // format of response to AA 02 command
		UInt16 checksum;
		char unknown1[4];
		UInt32 eout;
		char unknown2[6];
	} f2;
	struct {  // format of response to 99 01 command
		UInt16 checksum;
		UInt16 serialnumber;
		char unknown2[12];  // might be revision codes
	} f3;
} AnswerBuf;

// This holds settings and data for NVRAM that survive
// the termination of the program.
typedef struct {
    int interval;		// sample interval in units below
    int units;			// units of the interval 0=seconds, 1=minutes, ...
    UInt32 lastsampletime;  // time of last sample, in seconds since 1904
    AnswerBuf s1;
    AnswerBuf s2;
    AnswerBuf s3;
} NvramType;

NvramType Nvram;

static LocalID g_dbID = 0;		/* nonzero if dababase is open */
static DmOpenRef g_dbR;		/* DmOpenRef to open database, 0 no open */

// This is a buffer for one uncompressed record
// Either read from disk, or to be written to disk
Buffer WorkingBuffer;
char WorkingArray[MAXTEXTSIZE+1];

// pointers to various controls - these do not change during program run
static ListPtr g_list;
static ControlPtr g_popup;      // file open trigger

static UInt16 PortID = 0;  // Stores handle for serial port

/***********************************************************************
 *
 *	Internal Functions
 *
 ***********************************************************************/
// A sound only dogs can hear - used for delay
static SndCommandType NilSound = {sndCmdFreqDurationAmp, 0, 20940, 0, 1};

// Delay loop
// input: t - time to wait (milliseconds)
// Granularity is not to 1 mS.
// Use sparingly, stops all other processing
void stall(UInt32 t)
{
 // The following would cause crashes with high compiler optimization
 // SysTaskDelay(t * SysTicksPerSecond() / 1000);

 NilSound.param2 = t;  // silent sound duration in milliseconds
 SndDoCmd(NULL, &NilSound, false);
}

static SndCommandType BeepSound = {sndCmdFreqDurationAmp, 0, 1047, 50, sndMaxAmp};

// A brief Beep, if enabled.
// otherwise, do an equivalent silent delay
void beep()
{
   // Can't just set volume.  If zero, lose delay.
   // Volume of 1 is still pretty loud on some devices.
      SndDoCmd(NULL, &BeepSound, false);
}

// This puts up a message in the status area at the
// top right of the screen, above the bar.
// Inputs: s - string to put up
//         blink - true if message should flicker to get attention
void StatusMsg(char *s, Boolean blink)
{
   UInt16 len;
   RectangleType EraseRect = {topLeft: {POS_STATUS, 0}, extent: {160 - POS_STATUS, 11}};

   // erase old string
   // WinEraseChars is actually slower than blanking the whole rectangle
   //   WinEraseChars(OldString, StrLen(OldString), POS_STATUS, 0);
   WinEraseRectangle(&EraseRect, 0);

   // put up new string
   len = StrLen(s);
   WinDrawChars(s, len, POS_STATUS, 0);

   if (blink) {
      stall(200);
      WinEraseRectangle(&EraseRect, 0);
      beep();
      WinDrawChars(s, len, POS_STATUS, 0);
      stall(200);
      WinEraseRectangle(&EraseRect, 0);
      beep();
      WinDrawChars(s, len, POS_STATUS, 0);
      stall(200);
   }
}

// Print an alert with a number.
// Main use is for debugging.
// Moved here, used to be in alertnum.c
void AlertNum(char *str, Int32 num)
{
   char buf[16];
   StrPrintF(buf, "%ld", num);
   FrmCustomAlert(alertInfo, str, " ", buf);
}

// Create a brand new DOC file
// Returns non-zero if unable to create, including if the file already
// exists
static Err CreateDoc(char *name, UInt16 version)
{
   LocalID lid;
   Err err;

   err = DmCreateDatabase(0, name, DocCreator, DocTypeID, false);

   if (err != 0) {
      StrPrintF(OutString, "create err 0%d", err);
      StatusMsg(OutString, true);  // can't create database
      return err;
   }

   lid = DmFindDatabase(0, name);
   if (lid == 0) {
      StatusMsg("create err 10", true);  // can't find database
      return err;
   }

   g_dbID = lid;

   DOC_Values.reserved1 = 0;
   DOC_Values.doc_size = 0;
   DOC_Values.count = 0;
   DOC_Values.rec_size = MAXTEXTSIZE;
   DOC_Values.position = 0;
   DOC_Values.version = version;

   // Must do this before any text written to the file;
   // this creates the header.
   WriteHeader(g_dbID);

   g_dbID = 0;  // Just insurance.  All Write routines close database, anyway

   return err;
}

// Many of these functions were leveraged from my program Zdocj,
// which is also GPL open source.  - Roderick.
// Open a database for reading.
// The first record of a DOC database contains standard information;
// this is read into a global structure.
// input: name - pointer to string with the name of desired database
// output: g_dbR set to handle of opened database
//         g_dbID set to local ID of database
// return: 0 on success
static Err OpenReadDatabase(char *name)
{
   MemHandle recHand;
   MemPtr recPtr;
   UInt16 maxrec;
   LocalID lid;

   lid = DmFindDatabase(0, name);
   if (lid == 0) {
      return 1;
   }
   g_dbR = DmOpenDatabase(0, lid, dmModeReadOnly);
   if (g_dbR == 0) {
      return 2;
   }
   g_dbID = lid;

   recHand = DmQueryRecord(g_dbR, 0);
   if (recHand == 0) {
      DmCloseDatabase(g_dbR);
      return 3;
   }
   recPtr = MemHandleLock(recHand);
   MemMove(&DOC_Values, recPtr, sizeof DOC_Values);
   MemPtrUnlock(recPtr);
   // DmReleaseRecord(g_dbR, 0, (Boolean) false);  Not needed, we only did a DmQueryRecord

   maxrec = DmNumRecords(g_dbR);
   if (maxrec <= DOC_Values.count) {
      DOC_Values.count = maxrec - 1;
      StatusMsg("r-count fixed", true);
   }

   // Check for versions which we can't fully handle.
   // Encryption (version 3) NOT handled.
   if (DOC_Values.version == 0 || DOC_Values.version > 2) {
      StrPrintF(OutString, "weird ver: %u", DOC_Values.version);
      StatusMsg(OutString, true);
   }

   return 0;
}

static MemPtr GetObjectPtr(UInt16 objID)
{
   FormPtr form;
   form = FrmGetActiveForm();
   return FrmGetObjectPtr(form, FrmGetObjectIndex(form, objID));
}

static FieldPtr SetFieldTextFromHandle(FieldPtr fldP, MemHandle txtH)
{
   MemHandle oldTxtH;
   oldTxtH = FldGetTextHandle(fldP);

   FldSetTextHandle(fldP, txtH);
   // free the handle AFTER we call FldSetTextHandle().
   if (oldTxtH) {
      MemHandleFree(oldTxtH);
   }
   return fldP;
}

static FieldPtr SetFieldTextFromStr(FieldPtr fieldP, char *strP, Int32 size)
{
   MemHandle txtH;
   MemPtr textPtr;
   Err err;
   size = size + 1;		/* add '\0' size */
   txtH = MemHandleNew(size);
   if (!txtH) {
      StatusMsg("alloc fail 21", true);
      return NULL;
   }
   textPtr = MemHandleLock(txtH);
   if (textPtr == 0) {
      StatusMsg("lock fail 22", true);
      return NULL;
   }
   
   MemMove(textPtr, strP, size);
   err = MemPtrUnlock(textPtr);
   if (err != 0) {
      StatusMsg("lock fail 23", true);
      return NULL;
   }
   return SetFieldTextFromHandle(fieldP, txtH);
}

// Input is the time in seconds in palm time (since 1904).
// The time is formatted and set into the fields defined by DateFieldID and
// TimeFieldID.
// As a side effect, the formatted text is also put into the output buffer
// intended for the log file.
static void TimeToFields(UInt16 DateFieldID, UInt16 TimeFieldID, UInt32 seconds)
{
	DateTimeType dt;
	char TempString[11];
	
	TimSecondsToDateTime(seconds, &dt);
	StrPrintF(TempString, "%04d-%02d-%02d", dt.year, dt.month, dt.day);
    StrCopy(WorkingArray, TempString);
	SetFieldTextFromStr(GetObjectPtr(DateFieldID), TempString, StrLen(TempString));
	StrPrintF(TempString, "%02d:%02d:%02d", dt.hour, dt.minute, dt.second);
    StrCat(WorkingArray, ",");
    StrCat(WorkingArray, TempString);
	SetFieldTextFromStr(GetObjectPtr(TimeFieldID), TempString, StrLen(TempString));
}

// Forward declaration
static void DoSample();

// This is called when an alarm goes off.
static void AlarmProc(UInt16 almProcCmd, SysAlarmTriggeredParamType *paramP)
{
	DoSample();
}

static void BlankNextSample()
{
	SetFieldTextFromStr(GetObjectPtr(fieldID_nextdate), " ", 1);
	SetFieldTextFromStr(GetObjectPtr(fieldID_nexttime), " ", 1);		
}

static void ScheduleNextSample()
{
	UInt32 BaseSeconds, Now, ScheduleSeconds;  // times since 1904
	UInt32 speriod;
	DateTimeType dt;
	
	Nvram.interval = StrAToI(FldGetTextPtr(GetObjectPtr(fieldID_sampleinterval)));
	speriod = Nvram.interval;
	Nvram.units = LstGetSelection(g_list);          
	switch (Nvram.units) {
		case 1: speriod *= 60;  // scale by minutes
		        break;
		case 2: speriod *= 3600;  // scale by hours
		        break;
		case 3: speriod *= 86400L;  // scale by days
		        break;
		default: break;  // anything else, assume to be seconds, no scale
	}
	// Schedule at the next round period of time, starting from the beginning
	// of the day.  For example, if the time is now 4:22:36 on 2007-9-4,
	// the next scheduled time is 4:22:45 if the period is 15 seconds, or
	// 6:00:00 if the period is 6 hours, or 0:00:00 on 2007-9-5 if the period
	// is 1 day
	if (speriod > 0) {
		Now = TimGetSeconds();
		TimSecondsToDateTime(Now, &dt);
		dt.hour = 0;
		dt.minute = 0;
		dt.second = 0;
		BaseSeconds = TimDateTimeToSeconds(&dt);
		ScheduleSeconds = (((Now - BaseSeconds) / speriod) + 1) * speriod + BaseSeconds;
		TimeToFields(fieldID_nextdate, fieldID_nexttime, ScheduleSeconds);
		AlmSetProcAlarm(AlarmProc, 0, ScheduleSeconds);		
	} else {
		AlertNum("Sampling interval must be greater than ", 0);
		BlankNextSample();
	}
}

// Display the sample stored in the "Nvram" structure in
// appropriate fields on the screen
// If WriteLogFile is true, also send the sample to the log file
// Custom: this routine would be modified if a different inveter is used.
static void DisplaySample(Boolean WriteLogFile)
{
	char OutString[10];
	UInt32 e10;  // % efficiency * 10

    // Timestamp automatically fills WorkingArray
	TimeToFields(fieldID_lastdate, fieldID_lasttime, Nvram.lastsampletime);
    // Total kWh out
	StrIToA(OutString, Nvram.s2.f2.eout);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_eout), OutString, StrLen(OutString));

	// DC output voltage (Utility voltage)
	StrPrintF(OutString, "%3ld.%01ld", (Int32) Nvram.s1.f1.vout * 10 / 552, (Int32) Nvram.s1.f1.vout * 100 / 552 % 10);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_vout), OutString, 5);
	// DC input current
	StrPrintF(OutString, "%2ld.%01ld", (Int32) Nvram.s1.f1.iin / 1246, (Int32) Nvram.s1.f1.iin * 10 / 1246 % 10);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_iin), OutString, 4);
	// DC input voltage
	StrPrintF(OutString, "%3ld.%01ld", (Int32) Nvram.s1.f1.vin * 10 / 647, (Int32) Nvram.s1.f1.vin * 100 / 647 % 10);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_vin), OutString, 5);
	// Power out
	StrIToA(OutString, Nvram.s1.f1.pout * 100L / 273);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_pout), OutString, StrLen(OutString));
    // Heat Sink Temperature
    // Extra work because temperature could possibly be negative
    if (Nvram.s1.f1.temp >= 0)
	    StrPrintF(OutString, "%3d.%01d", Nvram.s1.f1.temp / 100, Nvram.s1.f1.temp /10 % 10);
    else
    	StrPrintF(OutString, "-%2d.%01d", (-Nvram.s1.f1.temp) / 100, (-Nvram.s1.f1.temp) /10 % 10);
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_temp), OutString, 5);

    // Compute efficiency = Pout / (Vin * Iin)
    // Avoid divide by 0 if there is no power input
    // The case of no power output is handled by the else clause (e10 = 0)
    if (Nvram.s1.f1.iin == 0 || Nvram.s1.f1.vin == 0)
    	StrCopy(OutString, "n/a");
    else {
        e10 = 29800000L / Nvram.s1.f1.vin * Nvram.s1.f1.pout / Nvram.s1.f1.iin;
	    StrPrintF(OutString, "%ld.%ld", e10 / 10, e10 % 10);
    }
	StrCat(WorkingArray, ",");
	StrCat(WorkingArray, OutString);
	SetFieldTextFromStr(GetObjectPtr(fieldID_efficiency), OutString, 4);

    StrCat(WorkingArray, "\n");

	// Append new record to log file under most conditions
	if (WriteLogFile) {
		if (DOC_Values.count < 65535) {
		    WorkingBuffer.buf = WorkingArray;
		    WorkingBuffer.len = StrLen(WorkingArray);
		    WriteRec(g_dbID, true, DOC_Values.count+1, &WorkingBuffer);
			DOC_Values.doc_size += WorkingBuffer.len;
		} else {
			StatusMsg("LOGFILE FULL!", false);
		}
	}
}

// define this to spoof data to the input buffers, for demo purposes
// #define SPOOF

// define this to log raw data also
// #define LOGRAW

// Set to true to repeatedly have a new sampled scheduled
// after the present one is taken
static Boolean ContinuousSampling = false;

// Take the actual sample.
// This must be customized to the particular device from which we are
// gathering data.
// 1.  Gather, and process data.  Store in "Nvram" structure.
// 2.  Put a time stamp on the data.
// 3.  Display data from Nvram structure onto screen.
// 4.  Append the data onto the log file.
// 5.  If appropriate, schedule the next sampling time.
// Custom: this routine would be modified if a different inveter is used.
static void DoSample()
{
    UInt32 SerialFlags;
    Int16 SfSize;
#ifndef SPOOF
	Err error;
#endif
	
#ifdef LOGRAW
int DebugI;
char DebugString[10];
#endif

    // Blank out any existing message before starting
    StatusMsg(" ", false);

	Nvram.lastsampletime = TimGetSeconds();  // get timestamp

#ifndef SPOOF

    // Open the serial port - fixed at 9600 baud for this inverter
    // Custom: change to match your inverter
    error = SrmOpen(serPortCradlePort, 9600, &PortID);
    if (error != 0) AlertNum("Warning: can't open serial port ", 0);

    // Custom: The PV Powered inverter uses RTS as a half-duplex
    // control, not a flow control.  Just turn it off forever.
    // If RTS is active, the inverter will not send to the Palm.
    SerialFlags = srmSettingsFlagBitsPerChar8 | srmSettingsFlagRTSInactive;
    SfSize = sizeof(SerialFlags);
    SrmControl(PortID, srmCtlSetFlags, &SerialFlags, &SfSize);

	// Custom: Take the actual sample.
	SrmReceiveFlush(PortID, 0);
	SrmSend(PortID, "\xAA\x01", 2, &error);  // Command: get info 1
	if (error != 0) {
		StrPrintF(WorkingArray, "Tx err %d", error);
		StatusMsg(WorkingArray, false);
		return;
	}
	// Get 16 bytes response, 1 second timeout if no answer
	SrmReceive(PortID, Nvram.s1.buf, 16, SysTicksPerSecond(), &error);
	if (error != 0) {
		StrPrintF(WorkingArray, "Rx err %d", error);
		StatusMsg(WorkingArray, false);
		return;
	}
	
#else

Nvram.s1.buf[0] = 0xFE;
Nvram.s1.buf[1] = 0x55;
Nvram.s1.buf[2] = 0x15;
Nvram.s1.buf[3] = 0x00;
Nvram.s1.buf[4] = 0x00;
Nvram.s1.buf[5] = 0x00;
Nvram.s1.buf[6] = 0x0B;
Nvram.s1.buf[7] = 0x18;
Nvram.s1.buf[8] = 0x9F;
Nvram.s1.buf[9] = 0x32;
Nvram.s1.buf[10] = 0x63;
Nvram.s1.buf[11] = 0x3B;
Nvram.s1.buf[12] = 0xC9;
Nvram.s1.buf[13] = 0x11;
Nvram.s1.buf[14] = 0x21;
Nvram.s1.buf[15] = 0x34;

#endif

#ifdef LOGRAW
// Log the raw bytes to the file
StrCopy(WorkingArray, "aa01: ");
for (DebugI=0; DebugI<16; DebugI++) {
	StrPrintF(DebugString, "%hx ", Nvram.s1.buf[DebugI]);
	StrCat(WorkingArray, &DebugString[2]);
}
StrCat(WorkingArray, "\n");
WorkingBuffer.buf = WorkingArray;
WorkingBuffer.len = StrLen(WorkingArray);
WriteRec(g_dbID, true, DOC_Values.count+1, &WorkingBuffer);
DOC_Values.doc_size += WorkingBuffer.len;
#endif

#ifndef SPOOF

	SrmReceiveFlush(PortID, 0);
	SrmSend(PortID, "\xAA\x02", 2, &error);  // Command: get info 2
	if (error != 0) {
		StrPrintF(WorkingArray, "Tx err %d", error);
		StatusMsg(WorkingArray, false);
		return;
	}
	// Get 16 bytes response, 1 second timeout if no answer
	SrmReceive(PortID, Nvram.s2.buf, 16, SysTicksPerSecond(), &error);
	if (error != 0) {
		StrPrintF(WorkingArray, "Rx err %d", error);
		StatusMsg(WorkingArray, false);
		return;
	}
	
#else

Nvram.s2.buf[0] = 0xFD;
Nvram.s2.buf[1] = 0x55;
Nvram.s2.buf[2] = 0x00;
Nvram.s2.buf[3] = 0x00;
Nvram.s2.buf[4] = 0x00;
Nvram.s2.buf[5] = 0x00;
Nvram.s2.buf[6] = 0x9C;
Nvram.s2.buf[7] = 0x1A;
Nvram.s2.buf[8] = 0x00;
Nvram.s2.buf[9] = 0x00;
Nvram.s2.buf[10] = 0x00;
Nvram.s2.buf[11] = 0x00;
Nvram.s2.buf[12] = 0x00;
Nvram.s2.buf[13] = 0x00;
Nvram.s2.buf[14] = 0x00;
Nvram.s2.buf[15] = 0x00;

#endif

#ifdef LOGRAW
// Log the raw bytes to the file
StrCopy(WorkingArray, "aa02: ");
for (DebugI=0; DebugI<16; DebugI++) {
	StrPrintF(DebugString, "%hx ", Nvram.s2.buf[DebugI]);
	StrCat(WorkingArray, &DebugString[2]);
}
StrCat(WorkingArray, "\n");
WorkingBuffer.buf = WorkingArray;
WorkingBuffer.len = StrLen(WorkingArray);
WriteRec(g_dbID, true, DOC_Values.count+1, &WorkingBuffer);
DOC_Values.doc_size += WorkingBuffer.len;
#endif

//  This only gets the unit's serial number, as far as I know.  Skip.
//	SrmReceiveFlush(PortID, 0);
//	SrmSend(PortID, "\x99\x01", 2, &error);  // Command: get info 3
//	if (error != 0) {
//		StrPrintF(WorkingArray, "Tx err %d", error);
//		StatusMsg(WorkingArray, false);
//		return;
//	}
//	// Get 16 bytes response, 1 second timeout if no answer
//	SrmReceive(PortID, Nvram.s3.buf, 16, SysTicksPerSecond(), &error);
//	if (error != 0) {
//		StrPrintF(WorkingArray, "Rx err %d", error);
//		StatusMsg(WorkingArray, false);
//		return;
//	}

	// Close the serial port - save power between samples
	SrmClose(PortID);

	// Fix up values to Motorola Big-Endian
    Nvram.s2.f2.eout = EndianSwap32(Nvram.s2.f2.eout);
	Nvram.s1.f1.vout = EndianSwap16(Nvram.s1.f1.vout);
	Nvram.s1.f1.iin  = EndianSwap16(Nvram.s1.f1.iin);
	Nvram.s1.f1.vin  = EndianSwap16(Nvram.s1.f1.vin);
	Nvram.s1.f1.pout = EndianSwap16(Nvram.s1.f1.pout);
	Nvram.s1.f1.temp = EndianSwap16(Nvram.s1.f1.temp);
	
	DisplaySample(true);  // show data on screen, and log to file
	
	// if running continuously, schedule next sample 
	if (ContinuousSampling)
	    ScheduleNextSample();
	else
		BlankNextSample();
	
	FrmDrawForm(FrmGetActiveForm());
}

static char *interval_list[N_INTERVALS] = {"seconds", "minutes", "hours", "days"};
static char triggerLabel[8];

// Things to do at startup, when the main form is opened
// This includes initializing the form with
// the most recently recorded values, if any
static void MainFormInit(FormPtr form)
{
	// Initialize the units for the sampling interval
    g_list = (ListPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, listID_interval));
    g_popup = (ControlPtr) FrmGetObjectPtr(form, FrmGetObjectIndex(form, popupID_interval));
    LstSetListChoices(g_list, interval_list, N_INTERVALS);
    LstSetSelection(g_list, Nvram.units);
    StrCopy(triggerLabel, LstGetSelectionText(g_list, Nvram.units));
    CategorySetTriggerLabel(g_popup, triggerLabel);
	StrIToA(WorkingArray, Nvram.interval);
	SetFieldTextFromStr(GetObjectPtr(fieldID_sampleinterval), WorkingArray, StrLen(WorkingArray));

    // Display the last sample, if any
    if (Nvram.lastsampletime != 0) {
        DisplaySample(false);
    }

}

/***********************************************************************
 *
 * FUNCTION:    MainFormHandleEvent
 *
 * DESCRIPTION: This routine is the event handler for the 
 *              "MainForm" of this application.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean MainFormHandleEvent(EventPtr event)
{
	Boolean 	handled = false;
	FormPtr 	pForm;

	switch (event->eType) {

    	case menuEvent:
			switch (event->data.menu.itemID) {
				case menuitemID_about:
				    FrmCustomAlert(alertID_about, pvm1Version, __DATE__, __TIME__);
				    handled = true;
				    break;
				case menuitemID_help:
				    FrmHelp(stringID_help);
				    handled = true;
				    break;
	  			default:
	    			break;
			}
			break;

		case frmOpenEvent:
			pForm = FrmGetActiveForm();
			MainFormInit(pForm);
			FrmDrawForm(pForm);
			handled = true;
			break;

	    case ctlSelectEvent:
	
	      // A control button was pressed and released.
	      switch (event->data.ctlEnter.controlID) {
	            
	         case buttonID_single:
	         	AlmSetProcAlarm(AlarmProc, 0, 0);  // cancel alarm
	            ContinuousSampling = false;
	         	DoSample();
	            handled = true;
	            break;
	            
	         case buttonID_start:
	            ContinuousSampling = true;
	            ScheduleNextSample();
                FrmDrawForm(FrmGetActiveForm());
	         	handled = true;
	         	break;
	         	
	         case buttonID_stop:
	         	AlmSetProcAlarm(AlarmProc, 0, 0);  // cancel alarm
	            ContinuousSampling = false;
                BlankNextSample();
                FrmDrawForm(FrmGetActiveForm());
	            handled = true;
	            break;
	            
	         default:
	            break;
	      }

		default:
			break;
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:    AppHandleEvent
 *
 * DESCRIPTION: This routine loads form resources and set the event
 *              handler for the form loaded.
 *
 * PARAMETERS:  event  - a pointer to an EventType structure
 *
 * RETURNED:    true if the event has handle and should not be passed
 *              to a higher level handler.
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Boolean AppHandleEvent(EventType* pEvent)
{
	UInt16 		formId;
	FormType* 	pForm;
	Boolean		handled = false;

	if (pEvent->eType == frmLoadEvent) {
		// Load the form resource.
		formId = pEvent->data.frmLoad.formID;
		
		pForm = FrmInitForm(formId);
		FrmSetActiveForm(pForm);

		// Set the event handler for the form.  The handler of the currently
		// active form is called by FrmHandleEvent each time is receives an
		// event.
		switch (formId) {
			case formID_pvm1:
				FrmSetEventHandler(pForm, MainFormHandleEvent);
				break;

			default:
				break;
		}
		handled = true;
	}
	
	return handled;
}


/***********************************************************************
 *
 * FUNCTION:     AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * PARAMETERS:   nothing
 *
 * RETURNED:     Err value errNone if nothing went wrong
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static Err AppStart(void)
{
	Err error;
    UInt16 size;
    Int16 ver;

	// Open the log file
	error = OpenReadDatabase(LOGFILENAME);
	// If it doesn't exist, create it.
	if (error != 0) {
        FrmCustomAlert(alertInfo, "Creating new logfile: ", LOGFILENAME, " ");
		error = CreateDoc(LOGFILENAME, 1);  // create uncompressed document
		ErrFatalDisplayIf(error != 0, "Unable to create logfile.");
		error = OpenReadDatabase(LOGFILENAME);
		ErrFatalDisplayIf(error != 0, "Unable to open logfile.");

		// Write first line of labels
		// Custom: this might change for a different model inverter
		StrCopy(WorkingArray, "Date,Time,Total Power kWH,Utility RMS Voltage,PV RMS Current,PV DC In Voltage,Power Out,Heat Sink Temperature,Efficiency\n");
        WorkingBuffer.buf = WorkingArray;
        WorkingBuffer.len = StrLen(WorkingArray);
        WriteRec(g_dbID, true, DOC_Values.count+1, &WorkingBuffer);
	    DOC_Values.doc_size += WorkingBuffer.len;
	}	

    // Get readings from last time
    size = sizeof(Nvram);
    ver =
      PrefGetAppPreferences(pvm1AppID, PREFID_BASE, &Nvram, &size, false);
    // if anything about retrieved preferences doesn't match this
    // build of the program, set defaults
    if (sizeof(Nvram) != size || ver != PREF_VER) {
    	Nvram.lastsampletime = 0;  // 0 = invalid sample
    	Nvram.interval = 6;  // default sample period 6
    	Nvram.units = 2;     // hours
        FrmHelp(stringID_help);  // very first time, show splash
    }

 	FrmGotoForm(formID_pvm1);
	return errNone;
}

/***********************************************************************
 *
 * FUNCTION:    AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppStop(void)
{
	// Stop next sample
    AlmSetProcAlarm(AlarmProc, 0, 0);  // cancel alarm if any
    
	// update the DOC header
    WriteHeader(g_dbID);
	
	// close the log file, if it is still open
	if (g_dbR != 0) DmCloseDatabase(g_dbR);		

	// remember value of sample interval
    Nvram.interval = StrAToI(FldGetTextPtr(GetObjectPtr(fieldID_sampleinterval)));
    Nvram.units = LstGetSelection(g_list);
	// save the last recorded settings
    PrefSetAppPreferences(pvm1AppID, PREFID_BASE, PREF_VER,
		&Nvram, sizeof(Nvram), false);

	// Close all the open forms.
	FrmCloseAllForms();
}

/***********************************************************************
 *
 * FUNCTION:    AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.  
 *
 * PARAMETERS:  nothing
 *
 * RETURNED:    nothing
 *
 * REVISION HISTORY:
 *
 *
 ***********************************************************************/
static void AppEventLoop(void)
{
	Err			error;
	EventType	event;

	do {
		EvtGetEvent(&event, evtWaitForever);

		if (SysHandleEvent(&event))
			continue;
			
		if (MenuHandleEvent(0, &event, &error))
			continue;
			
		if (AppHandleEvent(&event))
			continue;

		FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


/***********************************************************************
 *
 * FUNCTION:    PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:  cmd - word value specifying the launch code. 
 *              cmdPB - pointer to a structure that is associated with the launch code. 
 *              launchFlags -  word value providing extra information about the launch.
 * RETURNED:    Result of launch
 *
 * REVISION HISTORY: 
 *
 *
 ***********************************************************************/
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err error = errNone;

	switch (cmd) {
		case sysAppLaunchCmdNormalLaunch:
			if ((error = AppStart()) == 0) {			
				AppEventLoop();
				AppStop();
			}
			break;

		default:
			break;
	}
	
	return error;
}
